"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonTemplateParser = void 0;
const types_1 = require("./types");
const errors_1 = require("./errors");
const constants_1 = require("./constants");
const utils_1 = require("./utils");
const EMPTY_EXPR = { type: types_1.SyntaxType.EMPTY };
class JsonTemplateParser {
    constructor(lexer) {
        this.lexer = lexer;
    }
    parse() {
        this.lexer.init();
        return this.parseStatementsExpr();
    }
    parseEndOfStatement(blockEnd) {
        if (this.lexer.matchEOT() || this.lexer.match(blockEnd)) {
            return;
        }
        if (this.lexer.match(';')) {
            this.lexer.lex();
            return;
        }
        const currIdx = this.lexer.currentIndex();
        const nextTokenStart = this.lexer.lookahead().range[0];
        const code = this.lexer.getCodeChars(currIdx, nextTokenStart);
        if (!code.includes('\n')) {
            this.lexer.throwUnexpectedToken();
        }
    }
    parseStatements(blockEnd) {
        let statements = [];
        while (!this.lexer.matchEOT() && !this.lexer.match(blockEnd)) {
            statements.push(this.parseStatementExpr());
            this.parseEndOfStatement(blockEnd);
        }
        return statements;
    }
    parseStatementsExpr(blockEnd) {
        return {
            type: types_1.SyntaxType.STATEMENTS_EXPR,
            statements: this.parseStatements(blockEnd),
        };
    }
    parseStatementExpr() {
        return this.parseBaseExpr();
    }
    parseAssignmentExpr() {
        const expr = this.parseNextExpr(types_1.OperatorType.ASSIGNMENT);
        if (expr.type === types_1.SyntaxType.PATH && this.lexer.match('=')) {
            this.lexer.lex();
            return {
                type: types_1.SyntaxType.ASSIGNMENT_EXPR,
                value: this.parseBaseExpr(),
                path: expr,
            };
        }
        return expr;
    }
    parseBaseExpr() {
        return this.parseNextExpr(types_1.OperatorType.BASE);
    }
    parseNextExpr(currentOperation) {
        switch (currentOperation) {
            case types_1.OperatorType.CONDITIONAL:
                return this.parseAssignmentExpr();
            case types_1.OperatorType.ASSIGNMENT:
                return this.parseCoalescingExpr();
            case types_1.OperatorType.COALESCING:
                return this.parseLogicalORExpr();
            case types_1.OperatorType.OR:
                return this.parseLogicalANDExpr();
            case types_1.OperatorType.AND:
                return this.parseEqualityExpr();
            case types_1.OperatorType.EQUALITY:
                return this.parseRelationalExpr();
            case types_1.OperatorType.RELATIONAL:
                return this.parseShiftExpr();
            case types_1.OperatorType.SHIFT:
                return this.parseAdditiveExpr();
            case types_1.OperatorType.ADDITION:
                return this.parseMultiplicativeExpr();
            case types_1.OperatorType.MULTIPLICATION:
                return this.parsePowerExpr();
            case types_1.OperatorType.POWER:
                return this.parseUnaryExpr();
            case types_1.OperatorType.UNARY:
                return this.parsePathAfterExpr();
            default:
                return this.parseConditionalExpr();
        }
    }
    parseToArrayExpr() {
        this.lexer.lex();
        this.lexer.lex();
        return { type: types_1.SyntaxType.TO_ARRAY };
    }
    parsePathPart() {
        if (this.lexer.match('.') && this.lexer.match('(', 1)) {
            this.lexer.lex();
            return this.parseBlockExpr();
        }
        else if (this.lexer.match('(')) {
            return this.parseFunctionCallExpr();
        }
        else if (this.lexer.matchPathPartSelector()) {
            return this.parseSelector();
        }
        else if (this.lexer.match('[') && this.lexer.match(']', 1)) {
            return this.parseToArrayExpr();
        }
        else if (this.lexer.match('[')) {
            return this.parseArrayFiltersExpr();
        }
        else if (this.lexer.match('{')) {
            return this.parseObjectFiltersExpr();
        }
    }
    parsePathParts() {
        let parts = [];
        let newParts;
        while ((newParts = this.parsePathPart())) {
            newParts = utils_1.CommonUtils.toArray(newParts);
            parts.push(...newParts);
            if (newParts[0].type === types_1.SyntaxType.FUNCTION_CALL_EXPR) {
                break;
            }
        }
        return JsonTemplateParser.combinePathParts(parts);
    }
    static prependFunctionID(prefix, id) {
        return id ? prefix + '.' + id : prefix;
    }
    static combinePathParts(parts) {
        if (parts.length < 2) {
            return parts;
        }
        let newParts = [];
        for (let i = 0; i < parts.length; i++) {
            let expr = parts[i];
            if (i !== parts.length - 1) {
                if (expr.type === types_1.SyntaxType.TO_ARRAY) {
                    continue;
                }
                if (expr.type === types_1.SyntaxType.SELECTOR && expr.selector === '.') {
                    const selectorExpr = expr;
                    if (!selectorExpr.prop) {
                        continue;
                    }
                    else if (!selectorExpr.contextVar &&
                        selectorExpr.prop?.type === types_1.TokenType.ID &&
                        parts[i + 1].type === types_1.SyntaxType.FUNCTION_CALL_EXPR) {
                        expr = parts[i + 1];
                        expr.id = this.prependFunctionID(selectorExpr.prop.value, expr.id);
                        expr.dot = true;
                        i++;
                    }
                }
            }
            newParts.push(expr);
        }
        if (newParts.length < parts.length) {
            newParts = this.combinePathParts(newParts);
        }
        return newParts;
    }
    static convertToFunctionCallExpr(expr) {
        const fnExpr = expr.parts.pop();
        if (!expr.parts.length && expr.root && typeof expr.root !== 'object') {
            fnExpr.id = this.prependFunctionID(expr.root, fnExpr.id);
            fnExpr.dot = false;
        }
        else {
            fnExpr.object = expr;
        }
        return fnExpr;
    }
    parsePathRoot(root) {
        if (root) {
            return root;
        }
        if (this.lexer.match('^')) {
            this.lexer.lex();
            return constants_1.DATA_PARAM_KEY;
        }
        else if (this.lexer.matchID()) {
            return this.lexer.value();
        }
    }
    parsePath(root) {
        let expr = {
            type: types_1.SyntaxType.PATH,
            root: this.parsePathRoot(root),
            parts: this.parsePathParts(),
        };
        if (!expr.parts.length) {
            return expr;
        }
        JsonTemplateParser.setSubpath(expr.parts);
        const shouldConvertAsBlock = JsonTemplateParser.pathContainsVariables(expr.parts);
        const lastPart = utils_1.CommonUtils.getLastElement(expr.parts);
        if (lastPart.type === types_1.SyntaxType.TO_ARRAY) {
            expr.parts.pop();
            expr.toArray = true;
        }
        else if (lastPart.type === types_1.SyntaxType.FUNCTION_CALL_EXPR) {
            expr = JsonTemplateParser.convertToFunctionCallExpr(expr);
        }
        return shouldConvertAsBlock ? JsonTemplateParser.convertToBlockExpr(expr) : expr;
    }
    parseContextVariable(expected) {
        if (this.lexer.match(expected)) {
            this.lexer.lex();
            if (!this.lexer.matchID()) {
                this.lexer.throwUnexpectedToken();
            }
            return this.lexer.value();
        }
    }
    parseSelector() {
        let selector = this.lexer.value();
        let prop;
        let context;
        if (this.lexer.match('*')) {
            prop = this.lexer.lex();
        }
        if (this.lexer.matchID() || this.lexer.matchTokenType(types_1.TokenType.STR)) {
            prop = this.lexer.lex();
            while (this.lexer.match('@') || this.lexer.match('#')) {
                context = context || {};
                context.item = context.item || this.parseContextVariable('@');
                context.index = context.index || this.parseContextVariable('#');
            }
        }
        return {
            type: types_1.SyntaxType.SELECTOR,
            selector: selector,
            prop,
            context,
        };
    }
    parseRangeFilterExpr() {
        if (this.lexer.match(':')) {
            this.lexer.lex();
            return {
                type: types_1.SyntaxType.RANGE_FILTER_EXPR,
                toIdx: this.parseBaseExpr(),
            };
        }
        let fromExpr = this.parseBaseExpr();
        if (this.lexer.match(':')) {
            this.lexer.lex();
            if (this.lexer.match(']')) {
                return {
                    type: types_1.SyntaxType.RANGE_FILTER_EXPR,
                    fromIdx: fromExpr,
                };
            }
            return {
                type: types_1.SyntaxType.RANGE_FILTER_EXPR,
                fromIdx: fromExpr,
                toIdx: this.parseBaseExpr(),
            };
        }
        return fromExpr;
    }
    parseArrayIndexFilterExpr(expr) {
        const parts = [];
        if (expr) {
            parts.push(expr);
            if (!this.lexer.match(']')) {
                this.lexer.expect(',');
            }
        }
        return {
            type: types_1.SyntaxType.ARRAY_INDEX_FILTER_EXPR,
            indexes: {
                type: types_1.SyntaxType.ARRAY_EXPR,
                elements: [
                    ...parts,
                    ...this.parseCommaSeparatedElements(']', () => this.parseSpreadExpr()),
                ],
            },
        };
    }
    parseArrayFilterExpr() {
        if (this.lexer.matchSpread()) {
            return this.parseArrayIndexFilterExpr();
        }
        const expr = this.parseRangeFilterExpr();
        if (expr.type === types_1.SyntaxType.RANGE_FILTER_EXPR) {
            return expr;
        }
        return this.parseArrayIndexFilterExpr(expr);
    }
    parseObjectFilter() {
        let exclude = false;
        if (this.lexer.match('~')) {
            this.lexer.lex();
            exclude = true;
        }
        // excluding is applicaple only for index filters
        if (exclude || this.lexer.match('[')) {
            return {
                type: types_1.SyntaxType.OBJECT_INDEX_FILTER_EXPR,
                indexes: this.parseArrayExpr(),
                exclude,
            };
        }
        return {
            type: types_1.SyntaxType.OBJECT_FILTER_EXPR,
            filter: this.parseBaseExpr(),
        };
    }
    combineObjectFilters(objectFilters) {
        if (objectFilters.length <= 1) {
            return objectFilters;
        }
        const expr1 = objectFilters.shift();
        const expr2 = this.combineObjectFilters(objectFilters);
        return [
            {
                type: types_1.SyntaxType.OBJECT_FILTER_EXPR,
                filter: {
                    type: types_1.SyntaxType.LOGICAL_AND_EXPR,
                    op: '&&',
                    args: [expr1.filter, expr2[0].filter],
                },
            },
        ];
    }
    parseObjectFiltersExpr() {
        const objectFilters = [];
        const indexFilters = [];
        while (this.lexer.match('{')) {
            this.lexer.expect('{');
            const expr = this.parseObjectFilter();
            if (expr.type === types_1.SyntaxType.OBJECT_INDEX_FILTER_EXPR) {
                indexFilters.push(expr);
            }
            else {
                objectFilters.push(expr);
            }
            this.lexer.expect('}');
            if (this.lexer.match('.') && this.lexer.match('{', 1)) {
                this.lexer.lex();
            }
        }
        return [...this.combineObjectFilters(objectFilters), ...indexFilters];
    }
    parseConditionalExpr() {
        const ifExpr = this.parseNextExpr(types_1.OperatorType.CONDITIONAL);
        if (this.lexer.match('?')) {
            this.lexer.lex();
            const thenExpr = this.parseConditionalExpr();
            if (this.lexer.match(':')) {
                this.lexer.lex();
                const elseExpr = this.parseConditionalExpr();
                return {
                    type: types_1.SyntaxType.CONDITIONAL_EXPR,
                    if: ifExpr,
                    then: thenExpr,
                    else: elseExpr,
                };
            }
            return {
                type: types_1.SyntaxType.CONDITIONAL_EXPR,
                if: ifExpr,
                then: thenExpr,
                else: {
                    type: types_1.SyntaxType.LITERAL,
                    tokenType: types_1.TokenType.UNDEFINED,
                },
            };
        }
        return ifExpr;
    }
    parseArrayFiltersExpr() {
        const filters = [];
        while (this.lexer.match('[') && !this.lexer.match(']', 1)) {
            this.lexer.expect('[');
            filters.push(this.parseArrayFilterExpr());
            this.lexer.expect(']');
            if (this.lexer.match('.') && this.lexer.match('[', 1)) {
                this.lexer.lex();
            }
        }
        return {
            type: types_1.SyntaxType.ARRAY_FILTER_EXPR,
            filters,
        };
    }
    parseCoalescingExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.COALESCING);
        if (this.lexer.match('??')) {
            return {
                type: types_1.SyntaxType.LOGICAL_COALESCE_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseCoalescingExpr()],
            };
        }
        return expr;
    }
    parseLogicalORExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.OR);
        if (this.lexer.match('||')) {
            return {
                type: types_1.SyntaxType.LOGICAL_OR_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseLogicalORExpr()],
            };
        }
        return expr;
    }
    parseLogicalANDExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.AND);
        if (this.lexer.match('&&')) {
            return {
                type: types_1.SyntaxType.LOGICAL_AND_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseLogicalANDExpr()],
            };
        }
        return expr;
    }
    parseEqualityExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.EQUALITY);
        if (this.lexer.match('==') ||
            this.lexer.match('!=') ||
            this.lexer.match('===') ||
            this.lexer.match('!==') ||
            this.lexer.match('^==') ||
            this.lexer.match('==^') ||
            this.lexer.match('^=') ||
            this.lexer.match('=^') ||
            this.lexer.match('$==') ||
            this.lexer.match('==$') ||
            this.lexer.match('$=') ||
            this.lexer.match('=$') ||
            this.lexer.match('*==') ||
            this.lexer.match('==*') ||
            this.lexer.match('*=') ||
            this.lexer.match('=*')) {
            return {
                type: types_1.SyntaxType.COMPARISON_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseEqualityExpr()],
            };
        }
        return expr;
    }
    parseRelationalExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.RELATIONAL);
        if (this.lexer.match('<') ||
            this.lexer.match('>') ||
            this.lexer.match('<=') ||
            this.lexer.match('>=') ||
            this.lexer.matchIN()) {
            return {
                type: this.lexer.matchIN() ? types_1.SyntaxType.IN_EXPR : types_1.SyntaxType.COMPARISON_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseRelationalExpr()],
            };
        }
        return expr;
    }
    parseShiftExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.SHIFT);
        if (this.lexer.match('>>') || this.lexer.match('<<')) {
            return {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseShiftExpr()],
            };
        }
        return expr;
    }
    parseAdditiveExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.ADDITION);
        if (this.lexer.match('+') || this.lexer.match('-')) {
            return {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseAdditiveExpr()],
            };
        }
        return expr;
    }
    parseMultiplicativeExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.MULTIPLICATION);
        if (this.lexer.match('*') || this.lexer.match('/') || this.lexer.match('%')) {
            return {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parseMultiplicativeExpr()],
            };
        }
        return expr;
    }
    parsePowerExpr() {
        let expr = this.parseNextExpr(types_1.OperatorType.POWER);
        if (this.lexer.match('**')) {
            return {
                type: types_1.SyntaxType.MATH_EXPR,
                op: this.lexer.value(),
                args: [expr, this.parsePowerExpr()],
            };
        }
        return expr;
    }
    parseUnaryExpr() {
        if (this.lexer.match('!') ||
            this.lexer.match('-') ||
            this.lexer.matchTypeOf() ||
            this.lexer.matchAwait()) {
            return {
                type: types_1.SyntaxType.UNARY_EXPR,
                op: this.lexer.value(),
                arg: this.parseUnaryExpr(),
            };
        }
        return this.parseNextExpr(types_1.OperatorType.UNARY);
    }
    shouldSkipPathParsing(expr) {
        switch (expr.type) {
            case types_1.SyntaxType.DEFINTION_EXPR:
            case types_1.SyntaxType.ASSIGNMENT_EXPR:
            case types_1.SyntaxType.SPREAD_EXPR:
                return true;
            case types_1.SyntaxType.LITERAL:
            case types_1.SyntaxType.MATH_EXPR:
            case types_1.SyntaxType.COMPARISON_EXPR:
            case types_1.SyntaxType.ARRAY_EXPR:
            case types_1.SyntaxType.OBJECT_EXPR:
                if (this.lexer.match('(')) {
                    return true;
                }
                break;
            case types_1.SyntaxType.FUNCTION_EXPR:
                if (!this.lexer.match('(')) {
                    return true;
                }
                break;
        }
        return false;
    }
    parsePathAfterExpr() {
        let expr = this.parsePrimaryExpr();
        if (this.shouldSkipPathParsing(expr)) {
            return expr;
        }
        while (this.lexer.matchPathPartSelector() || this.lexer.match('[') || this.lexer.match('(')) {
            expr = this.parsePath(expr);
        }
        return expr;
    }
    parseLiteralExpr() {
        const token = this.lexer.lex();
        return {
            type: types_1.SyntaxType.LITERAL,
            value: token.value,
            tokenType: token.type,
        };
    }
    parseIDPath() {
        const idParts = [];
        while (this.lexer.matchID()) {
            idParts.push(this.lexer.value());
            if (this.lexer.match('.') && this.lexer.matchID(1)) {
                this.lexer.lex();
            }
        }
        if (!idParts.length) {
            this.lexer.throwUnexpectedToken();
        }
        return idParts.join('.');
    }
    parseObjectDefVars() {
        const vars = [];
        this.lexer.expect('{');
        while (!this.lexer.match('}')) {
            if (!this.lexer.matchID()) {
                throw new errors_1.JsosTemplateParserError('Invalid object vars at ' + this.lexer.getContext());
            }
            vars.push(this.lexer.value());
            if (!this.lexer.match('}')) {
                this.lexer.expect(',');
            }
        }
        this.lexer.expect('}');
        if (vars.length === 0) {
            throw new errors_1.JsosTemplateParserError('Empty object vars at ' + this.lexer.getContext());
        }
        return vars;
    }
    parseNormalDefVars() {
        const vars = [];
        if (!this.lexer.matchID()) {
            throw new errors_1.JsosTemplateParserError('Invalid normal vars at ' + this.lexer.getContext());
        }
        vars.push(this.lexer.value());
        return vars;
    }
    parseDefinitionExpr() {
        const definition = this.lexer.value();
        let fromObject = this.lexer.match('{');
        let vars = fromObject ? this.parseObjectDefVars() : this.parseNormalDefVars();
        this.lexer.expect('=');
        return {
            type: types_1.SyntaxType.DEFINTION_EXPR,
            value: this.parseBaseExpr(),
            vars,
            definition,
            fromObject,
        };
    }
    parseFunctionCallArgs() {
        this.lexer.expect('(');
        const args = this.parseCommaSeparatedElements(')', () => this.parseSpreadExpr());
        this.lexer.expect(')');
        return args;
    }
    parseFunctionCallExpr() {
        let id;
        if (this.lexer.matchNew()) {
            this.lexer.lex();
            id = 'new ' + this.parseIDPath();
        }
        return {
            type: types_1.SyntaxType.FUNCTION_CALL_EXPR,
            args: this.parseFunctionCallArgs(),
            id,
        };
    }
    parseFunctionDefinitionParam() {
        let spread = '';
        if (this.lexer.matchSpread()) {
            this.lexer.lex();
            spread = '...';
            // rest param should be last param.
            if (!this.lexer.match(')', 1)) {
                this.lexer.throwUnexpectedToken();
            }
        }
        if (!this.lexer.matchID()) {
            this.lexer.throwUnexpectedToken();
        }
        return `${spread}${this.lexer.value()}`;
    }
    parseFunctionDefinitionParams() {
        this.lexer.expect('(');
        const params = this.parseCommaSeparatedElements(')', () => this.parseFunctionDefinitionParam());
        this.lexer.expect(')');
        return params;
    }
    parseFunctionExpr(asyncFn = false) {
        this.lexer.lex();
        const params = this.parseFunctionDefinitionParams();
        this.lexer.expect('{');
        const statements = this.parseStatementsExpr('}');
        this.lexer.expect('}');
        return {
            type: types_1.SyntaxType.FUNCTION_EXPR,
            params,
            body: statements,
            async: asyncFn,
        };
    }
    parseObjectKeyExpr() {
        let key;
        if (this.lexer.match('[')) {
            this.lexer.lex();
            key = this.parseBaseExpr();
            this.lexer.expect(']');
        }
        else if (this.lexer.matchID()) {
            key = this.lexer.value();
        }
        else if (this.lexer.matchTokenType(types_1.TokenType.STR)) {
            key = this.parseLiteralExpr();
        }
        else {
            this.lexer.throwUnexpectedToken();
        }
        return key;
    }
    parseObjectPropExpr() {
        let key;
        if (!this.lexer.matchSpread()) {
            key = this.parseObjectKeyExpr();
            this.lexer.expect(':');
        }
        const value = this.parseSpreadExpr();
        return {
            type: types_1.SyntaxType.OBJECT_PROP_EXPR,
            key,
            value,
        };
    }
    parseObjectExpr() {
        this.lexer.expect('{');
        const props = this.parseCommaSeparatedElements('}', () => this.parseObjectPropExpr());
        this.lexer.expect('}');
        return {
            type: types_1.SyntaxType.OBJECT_EXPR,
            props,
        };
    }
    parseCommaSeparatedElements(blockEnd, parseFn) {
        const elements = [];
        while (!this.lexer.match(blockEnd)) {
            elements.push(parseFn());
            if (!this.lexer.match(blockEnd)) {
                this.lexer.expect(',');
            }
        }
        return elements;
    }
    parseSpreadExpr() {
        if (this.lexer.matchSpread()) {
            this.lexer.lex();
            return {
                type: types_1.SyntaxType.SPREAD_EXPR,
                value: this.parseBaseExpr(),
            };
        }
        return this.parseBaseExpr();
    }
    parseArrayExpr() {
        this.lexer.expect('[');
        const elements = this.parseCommaSeparatedElements(']', () => this.parseSpreadExpr());
        this.lexer.expect(']');
        return {
            type: types_1.SyntaxType.ARRAY_EXPR,
            elements,
        };
    }
    parseBlockExpr() {
        this.lexer.expect('(');
        let statements = this.parseStatements(')');
        this.lexer.expect(')');
        if (statements.length === 0) {
            return EMPTY_EXPR;
        }
        else if (statements.length === 1) {
            return statements[0];
        }
        return {
            type: types_1.SyntaxType.FUNCTION_EXPR,
            body: utils_1.CommonUtils.convertToStatementsExpr(...statements),
            block: true,
        };
    }
    parseAsyncFunctionExpr() {
        this.lexer.lex();
        if (this.lexer.matchFunction()) {
            return this.parseFunctionExpr(true);
        }
        else if (this.lexer.matchLambda()) {
            return this.parseLambdaExpr(true);
        }
        this.lexer.throwUnexpectedToken();
    }
    parseLambdaExpr(asyncFn = false) {
        this.lexer.lex();
        const expr = this.parseBaseExpr();
        return {
            type: types_1.SyntaxType.FUNCTION_EXPR,
            body: utils_1.CommonUtils.convertToStatementsExpr(expr),
            params: ['...args'],
            async: asyncFn,
        };
    }
    parsePrimaryExpr() {
        if (this.lexer.match(';')) {
            return EMPTY_EXPR;
        }
        if (this.lexer.matchTokenType(types_1.TokenType.LAMBDA_ARG)) {
            return {
                type: types_1.SyntaxType.LAMBDA_ARG,
                index: this.lexer.value(),
            };
        }
        if (this.lexer.matchNew()) {
            return this.parseFunctionCallExpr();
        }
        if (this.lexer.matchDefinition()) {
            return this.parseDefinitionExpr();
        }
        if (this.lexer.matchLambda()) {
            return this.parseLambdaExpr();
        }
        if (this.lexer.matchFunction()) {
            return this.parseFunctionExpr();
        }
        if (this.lexer.matchAsync()) {
            return this.parseAsyncFunctionExpr();
        }
        if (this.lexer.matchLiteral()) {
            return this.parseLiteralExpr();
        }
        if (this.lexer.match('{')) {
            return this.parseObjectExpr();
        }
        if (this.lexer.match('[')) {
            return this.parseArrayExpr();
        }
        if (this.lexer.matchPath()) {
            return this.parsePath();
        }
        if (this.lexer.match('(')) {
            return this.parseBlockExpr();
        }
        return this.lexer.throwUnexpectedToken();
    }
    static setSubpath(parts) {
        const remainingParts = parts.slice();
        while (remainingParts.length) {
            const part = remainingParts.shift();
            if (typeof part !== 'object') {
                continue;
            }
            if (part.type === types_1.SyntaxType.PATH && Array.isArray(part.parts)) {
                part.subPath = !part.root;
            }
            else {
                for (let key in part) {
                    if (Array.isArray(part[key])) {
                        remainingParts.push(...part[key].flat());
                    }
                    else if (typeof part[key] === 'object') {
                        remainingParts.push(part[key]);
                    }
                }
            }
        }
    }
    static pathContainsVariables(parts) {
        return parts
            .filter((part) => part.type === types_1.SyntaxType.SELECTOR)
            .map((part) => part)
            .some((part) => part.context?.index || part.context?.item);
    }
    static convertToBlockExpr(expr) {
        return {
            type: types_1.SyntaxType.FUNCTION_EXPR,
            block: true,
            body: utils_1.CommonUtils.convertToStatementsExpr(expr),
        };
    }
}
exports.JsonTemplateParser = JsonTemplateParser;
