"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.JsonTemplateTranslator = void 0;
const constants_1 = require("./constants");
const errors_1 = require("./errors");
const operators_1 = require("./operators");
const types_1 = require("./types");
const utils_1 = require("./utils");
class JsonTemplateTranslator {
    constructor(expr) {
        this.vars = [];
        this.lastVarId = 0;
        this.unusedVars = [];
        this.expr = expr;
    }
    init() {
        this.vars = [];
        this.lastVarId = 0;
        this.unusedVars = [];
    }
    acquireVar() {
        if (this.unusedVars.length) {
            return this.unusedVars.shift();
        }
        const varName = `${constants_1.VARS_PREFIX}${++this.lastVarId}`;
        this.vars.push(varName);
        return varName;
    }
    acquireVars(numVars = 1) {
        const vars = [];
        for (let i = 0; i < numVars; i++) {
            vars.push(this.acquireVar());
        }
        return vars;
    }
    releaseVars(...args) {
        let i = args.length;
        while (i--) {
            this.unusedVars.push(args[i]);
        }
    }
    translate(dest = constants_1.RESULT_KEY, ctx = constants_1.DATA_PARAM_KEY) {
        this.init();
        let code = [];
        const exprCode = this.translateExpr(this.expr, dest, ctx);
        code.push(`let ${dest};`);
        code.push(this.vars.map((elm) => `let ${elm};`).join(''));
        code.push(exprCode);
        code.push(`return ${dest};`);
        return code.join('');
    }
    translateExpr(expr, dest, ctx) {
        switch (expr.type) {
            case types_1.SyntaxType.STATEMENTS_EXPR:
                return this.translateStatementsExpr(expr, dest, ctx);
            case types_1.SyntaxType.PATH:
                return this.translatePath(expr, dest, ctx);
            case types_1.SyntaxType.IN_EXPR:
                return this.translateINExpr(expr, dest, ctx);
            case types_1.SyntaxType.COMPARISON_EXPR:
            case types_1.SyntaxType.MATH_EXPR:
                return this.translateBinaryExpr(expr, dest, ctx);
            case types_1.SyntaxType.LOGICAL_COALESCE_EXPR:
            case types_1.SyntaxType.LOGICAL_AND_EXPR:
            case types_1.SyntaxType.LOGICAL_OR_EXPR:
                return this.translateLogicalExpr(expr, dest, ctx);
            case types_1.SyntaxType.UNARY_EXPR:
                return this.translateUnaryExpr(expr, dest, ctx);
            case types_1.SyntaxType.LAMBDA_ARG:
                return this.translateLambdaArgExpr(expr, dest, ctx);
            case types_1.SyntaxType.SPREAD_EXPR:
                return this.translateSpreadExpr(expr, dest, ctx);
            case types_1.SyntaxType.LITERAL:
                return this.translateLiteralExpr(expr, dest, ctx);
            case types_1.SyntaxType.ARRAY_EXPR:
                return this.translateArrayExpr(expr, dest, ctx);
            case types_1.SyntaxType.OBJECT_EXPR:
                return this.translateObjectExpr(expr, dest, ctx);
            case types_1.SyntaxType.FUNCTION_EXPR:
                return this.translateFunctionExpr(expr, dest, ctx);
            case types_1.SyntaxType.FUNCTION_CALL_EXPR:
                return this.translateFunctionCallExpr(expr, dest, ctx);
            case types_1.SyntaxType.DEFINTION_EXPR:
                return this.translateDefinitionExpr(expr, dest, ctx);
            case types_1.SyntaxType.ASSIGNMENT_EXPR:
                return this.translateAssignmentExpr(expr, dest, ctx);
            case types_1.SyntaxType.OBJECT_FILTER_EXPR:
                return this.translateObjectFilterExpr(expr, dest, ctx);
            case types_1.SyntaxType.ARRAY_FILTER_EXPR:
                return this.translateArrayFilterExpr(expr, dest, ctx);
            case types_1.SyntaxType.OBJECT_INDEX_FILTER_EXPR:
                return this.translateIndexFilterExpr(expr, dest, ctx);
            case types_1.SyntaxType.SELECTOR:
                return this.translateSelector(expr, dest, ctx);
            case types_1.SyntaxType.CONDITIONAL_EXPR:
                return this.translateConditionalExpr(expr, dest, ctx);
            default:
                return '';
        }
    }
    translateConditionalExpr(expr, dest, ctx) {
        const code = [];
        const ifVar = this.acquireVar();
        const thenVar = this.acquireVar();
        const elseVar = this.acquireVar();
        code.push(this.translateExpr(expr.if, ifVar, ctx));
        code.push(`if(${ifVar}){`);
        code.push(this.translateExpr(expr.then, thenVar, ctx));
        code.push(`${dest} = ${thenVar};`);
        code.push('} else {');
        code.push(this.translateExpr(expr.else, elseVar, ctx));
        code.push(`${dest} = ${elseVar};`);
        code.push('}');
        this.releaseVars(elseVar);
        this.releaseVars(thenVar);
        this.releaseVars(ifVar);
        return code.join('');
    }
    translateLambdaArgExpr(expr, dest, ctx) {
        return `${dest} = args[${expr.index}];`;
    }
    translateSpreadExpr(expr, dest, ctx) {
        return this.translateExpr(expr.value, dest, ctx);
    }
    translatePathRoot(path, dest, ctx) {
        if (typeof path.root === 'object') {
            return this.translateExpr(path.root, dest, ctx);
        }
        else if (path.subPath && path.parts.length) {
            if (JsonTemplateTranslator.isSinglePropSelection(path.parts[0])) {
                const part = path.parts.shift();
                const propStr = utils_1.CommonUtils.escapeStr(part.prop?.value);
                return `${dest} = ${ctx}[${propStr}];`;
            }
        }
        return `${dest} = ${path.root || ctx};`;
    }
    translatePathContext(context, item, idx) {
        const code = [];
        if (context.item) {
            code.push(`let ${context.item} = ${item};`);
        }
        if (context.index) {
            code.push(`let ${context.index} = ${idx};`);
        }
        return code.join('');
    }
    prepareDataForPathPart(part, data) {
        const code = [];
        code.push(JsonTemplateTranslator.covertToArrayValue(data));
        if (JsonTemplateTranslator.isSinglePropSelection(part)) {
            const selector = part;
            const propStr = utils_1.CommonUtils.escapeStr(selector.prop?.value);
            code.push(`if(Object.prototype.hasOwnProperty.call(${data}, ${propStr})){`);
            code.push(`${data} = [${data}];`);
            code.push('}');
        }
        else if (JsonTemplateTranslator.isArrayFilterExpr(part)) {
            code.push(`${data} = [${data}];`);
        }
        return code.join('');
    }
    translatePathParts(expr, dest) {
        if (!expr.parts.length) {
            return '';
        }
        const parts = expr.parts;
        const code = [];
        const numParts = parts.length;
        const dataVars = this.acquireVars(numParts);
        const indexVars = this.acquireVars(numParts);
        const itemVars = this.acquireVars(numParts);
        const result = this.acquireVar();
        code.push(JsonTemplateTranslator.generateAssignmentCode(result, '[]'));
        code.push(JsonTemplateTranslator.generateAssignmentCode(dataVars[0], dest));
        for (let i = 0; i < numParts; i++) {
            const part = parts[i];
            const idx = indexVars[i];
            const item = itemVars[i];
            const data = dataVars[i];
            code.push(this.prepareDataForPathPart(part, data));
            code.push(`for(${idx}=0; ${idx}<${data}.length; ${idx}++) {`);
            code.push(`${item} = ${data}[${idx}];`);
            if (i > 0 && parts[i - 1].context) {
                code.push(this.translatePathContext(parts[i - 1].context, item, idx));
            }
            code.push(this.translateExpr(part, item, item));
            code.push(`if(!${item}) { continue; }`);
            if (i < numParts - 1) {
                code.push(JsonTemplateTranslator.generateAssignmentCode(dataVars[i + 1], item));
            }
            else {
                code.push(JsonTemplateTranslator.covertToArrayValue(item));
                code.push(`${result} = ${result}.concat(${item});`);
            }
        }
        for (let i = 0; i < numParts; i++) {
            code.push('}');
        }
        this.releaseVars(...indexVars);
        this.releaseVars(...itemVars);
        this.releaseVars(...dataVars);
        this.releaseVars(result);
        if (!expr.toArray) {
            code.push(result, '=', JsonTemplateTranslator.returnSingleValueIfSafe(result), ';');
        }
        code.push(JsonTemplateTranslator.generateAssignmentCode(dest, result));
        return code.join('');
    }
    translatePath(expr, dest, baseCtx) {
        const code = [];
        code.push(this.translatePathRoot(expr, dest, baseCtx));
        code.push(this.translatePathParts(expr, dest));
        if (expr.toArray && !expr.parts.length) {
            code.push(JsonTemplateTranslator.covertToArrayValue(dest));
        }
        return code.join('');
    }
    translateCurrentSelector(expr, dest, ctx) {
        const code = [];
        const prop = expr.prop?.value;
        if (prop === '*') {
            const valuesCode = JsonTemplateTranslator.returnObjectValues(ctx);
            code.push(`${dest} = ${valuesCode}.flat();`);
        }
        else if (prop) {
            const propStr = utils_1.CommonUtils.escapeStr(prop);
            code.push(`if(${ctx} && Object.prototype.hasOwnProperty.call(${ctx}, ${propStr})){`);
            code.push(`${dest}=${ctx}[${propStr}];`);
            code.push('} else {');
            code.push(`${dest} = undefined`);
            code.push('}');
        }
        return code.join('');
    }
    translateSelector(expr, dest, ctx) {
        if (expr.selector === '.') {
            return this.translateCurrentSelector(expr, dest, ctx);
        }
        return this.translateDescendantSelector(expr, dest, ctx);
    }
    translateDescendantSelector(expr, dest, baseCtx) {
        let code = [];
        const ctxs = this.acquireVar();
        const currCtx = this.acquireVar();
        const result = this.acquireVar();
        code.push(JsonTemplateTranslator.generateAssignmentCode(result, '[]'));
        const { prop } = expr;
        const propStr = utils_1.CommonUtils.escapeStr(prop?.value);
        code.push(`${ctxs}=[${baseCtx}];`);
        code.push(`while(${ctxs}.length > 0) {`);
        code.push(`${currCtx} = ${ctxs}.shift();`);
        code.push(`if(!${currCtx}){continue;}`);
        code.push(`if(Array.isArray(${currCtx})){`);
        code.push(`${ctxs} = ${ctxs}.concat(${currCtx});`);
        code.push('continue;');
        code.push('}');
        code.push(`if(typeof ${currCtx} === "object") {`);
        const valuesCode = JsonTemplateTranslator.returnObjectValues(currCtx);
        code.push(`${ctxs} = ${ctxs}.concat(${valuesCode});`);
        if (prop) {
            if (prop?.value === '*') {
                code.push(`${result} = ${result}.concat(${valuesCode});`);
            }
            else {
                code.push(`if(Object.prototype.hasOwnProperty.call(${currCtx}, ${propStr})){`);
                code.push(`${result}.push(${currCtx}[${propStr}]);`);
                code.push('}');
            }
        }
        code.push('}');
        if (!prop) {
            code.push(`${result}.push(${currCtx});`);
        }
        code.push('}');
        code.push(`${dest} = ${result}.flat();`);
        return code.join('');
    }
    translateFunctionExpr(expr, dest, ctx) {
        let code = [];
        const fnHead = expr.async ? 'async function' : 'function';
        code.push(dest, '=', fnHead, '(', (expr.params || []).join(','), '){');
        const fnTranslator = new JsonTemplateTranslator(expr.body);
        code.push(fnTranslator.translate(constants_1.FUNCTION_RESULT_KEY, ctx));
        code.push('}');
        if (expr.block) {
            code.push('()');
        }
        code.push(';');
        return code.join('');
    }
    getFunctionName(expr, ctx) {
        return expr.dot ? `${ctx}.${expr.id}` : expr.id || ctx;
    }
    translateFunctionCallExpr(expr, dest, ctx) {
        let code = [];
        const result = this.acquireVar();
        code.push(JsonTemplateTranslator.generateAssignmentCode(result, ctx));
        if (expr.object) {
            code.push(this.translateExpr(expr.object, result, ctx));
        }
        if (!expr.id) {
            code.push(JsonTemplateTranslator.convertToSingleValue(result));
        }
        const functionArgsStr = this.translateSpreadableExpressions(expr.args, result, code);
        code.push(dest, '=', this.getFunctionName(expr, result), '(', functionArgsStr, ');');
        return code.join('');
    }
    translateObjectExpr(expr, dest, ctx) {
        let code = [];
        const propExprs = [];
        const vars = [];
        for (let prop of expr.props) {
            const propParts = [];
            if (prop.key) {
                if (typeof prop.key !== 'string') {
                    const keyVar = this.acquireVar();
                    code.push(this.translateExpr(prop.key, keyVar, ctx));
                    propParts.push(`[${keyVar}]`);
                    vars.push(keyVar);
                }
                else {
                    propParts.push(prop.key);
                }
                propParts.push(':');
            }
            const valueVar = this.acquireVar();
            code.push(this.translateExpr(prop.value, valueVar, ctx));
            if (prop.value.type === types_1.SyntaxType.SPREAD_EXPR) {
                propParts.push('...');
            }
            propParts.push(valueVar);
            propExprs.push(propParts.join(''));
            vars.push(valueVar);
        }
        code.push(dest, '={', propExprs.join(','), '};');
        this.releaseVars(...vars);
        return code.join('');
    }
    translateSpreadableExpressions(items, ctx, code) {
        const vars = [];
        const itemParts = [];
        for (let item of items) {
            const varName = this.acquireVar();
            code.push(this.translateExpr(item, varName, ctx));
            itemParts.push(item.type === types_1.SyntaxType.SPREAD_EXPR ? `...${varName}` : varName);
            vars.push(varName);
        }
        this.releaseVars(...vars);
        return itemParts.join(',');
    }
    translateArrayExpr(expr, dest, ctx) {
        const code = [];
        const elementsStr = this.translateSpreadableExpressions(expr.elements, ctx, code);
        code.push(`${dest} = [${elementsStr}];`);
        return code.join('');
    }
    translateLiteralExpr(expr, dest, _ctx) {
        const literalCode = this.translateLiteral(expr.tokenType, expr.value);
        return JsonTemplateTranslator.generateAssignmentCode(dest, literalCode);
    }
    getSelectorAssignmentPart(expr) {
        if (!JsonTemplateTranslator.isValidSelectorForAssignment(expr)) {
            throw new errors_1.JsosTemplateTranslatorError('Invalid assignment path');
        }
        if (expr.prop?.type === types_1.TokenType.STR) {
            return `[${utils_1.CommonUtils.escapeStr(expr.prop?.value)}]`;
        }
        else {
            return `.${expr.prop?.value}`;
        }
    }
    getArrayIndexAssignmentPart(expr, code, ctx) {
        const parts = [];
        for (const filter of expr.filters) {
            if (filter.type === types_1.SyntaxType.RANGE_FILTER_EXPR) {
                throw new errors_1.JsosTemplateTranslatorError('Invalid assignment path');
            }
            if (filter.indexes.elements.length > 1) {
                throw new errors_1.JsosTemplateTranslatorError('Invalid assignment path');
            }
            const keyVar = this.acquireVar();
            code.push(this.translateExpr(filter.indexes.elements[0], keyVar, ctx));
            this.releaseVars(keyVar);
            parts.push(`[${keyVar}]`);
        }
        return parts.join('');
    }
    translateAssignmentExpr(expr, dest, ctx) {
        const code = [];
        const valueVar = this.acquireVar();
        code.push(this.translateExpr(expr.value, valueVar, ctx));
        const assignmentPathParts = [];
        const root = expr.path.root;
        if (!root || root === constants_1.DATA_PARAM_KEY || typeof root === 'object') {
            throw new errors_1.JsosTemplateTranslatorError('Invalid assignment path');
        }
        assignmentPathParts.push(root);
        for (let part of expr.path.parts) {
            switch (part.type) {
                case types_1.SyntaxType.SELECTOR:
                    assignmentPathParts.push(this.getSelectorAssignmentPart(part));
                    break;
                case types_1.SyntaxType.ARRAY_FILTER_EXPR:
                    assignmentPathParts.push(this.getArrayIndexAssignmentPart(part, code, ctx));
                    break;
                default:
                    throw new errors_1.JsosTemplateTranslatorError('Invalid assignment path');
            }
        }
        const assignmentPath = assignmentPathParts.join('');
        code.push(JsonTemplateTranslator.generateAssignmentCode(assignmentPath, valueVar));
        code.push(JsonTemplateTranslator.generateAssignmentCode(dest, valueVar));
        this.releaseVars(valueVar);
        return code.join('');
    }
    translateDefinitionVars(expr) {
        let vars = [expr.vars.join(',')];
        if (expr.fromObject) {
            vars.unshift('{');
            vars.push('}');
        }
        return vars.join('');
    }
    translateDefinitionExpr(expr, dest, ctx) {
        const code = [];
        const value = this.acquireVar();
        code.push(this.translateExpr(expr.value, value, ctx));
        const defVars = this.translateDefinitionVars(expr);
        code.push(`${expr.definition} ${defVars}=${value};`);
        code.push(JsonTemplateTranslator.generateAssignmentCode(dest, value));
        this.releaseVars(value);
        return code.join('');
    }
    translateStatementsExpr(expr, dest, ctx) {
        return this.translateStatements(expr.statements, dest, ctx);
    }
    translateStatements(statements, dest, ctx) {
        return statements.map((statement) => this.translateExpr(statement, dest, ctx)).join('');
    }
    getLogicalConditionCode(type, varName) {
        switch (type) {
            case types_1.SyntaxType.LOGICAL_COALESCE_EXPR:
                return `${varName} !== null && ${varName} !== undefined`;
            case types_1.SyntaxType.LOGICAL_OR_EXPR:
                return varName;
            default:
                return `!${varName}`;
        }
    }
    translateLogicalExpr(expr, dest, ctx) {
        const val1 = this.acquireVar();
        const code = [];
        code.push(this.translateExpr(expr.args[0], val1, ctx));
        const condition = this.getLogicalConditionCode(expr.type, val1);
        code.push(`if(${condition}) {`);
        code.push(JsonTemplateTranslator.generateAssignmentCode(dest, val1));
        code.push('} else {');
        const val2 = this.acquireVar();
        code.push(this.translateExpr(expr.args[1], val2, ctx));
        code.push(JsonTemplateTranslator.generateAssignmentCode(dest, val2));
        code.push('}');
        this.releaseVars(val1, val2);
        return code.join('');
    }
    translateINExpr(expr, dest, ctx) {
        const code = [];
        const val1 = this.acquireVar();
        const val2 = this.acquireVar();
        const resultVar = this.acquireVar();
        code.push(this.translateExpr(expr.args[0], val1, ctx));
        code.push(this.translateExpr(expr.args[1], val2, ctx));
        const inCode = `(Array.isArray(${val2}) ? ${val2}.includes(${val1}) : ${val1} in ${val2})`;
        code.push(JsonTemplateTranslator.generateAssignmentCode(resultVar, inCode));
        code.push(JsonTemplateTranslator.generateAssignmentCode(dest, resultVar));
        return code.join('');
    }
    translateLiteral(type, val) {
        if (type === types_1.TokenType.STR) {
            return utils_1.CommonUtils.escapeStr(val);
        }
        return String(val);
    }
    translateUnaryExpr(expr, dest, ctx) {
        const code = [];
        const val = this.acquireVar();
        code.push(this.translateExpr(expr.arg, val, ctx));
        code.push(`${dest} = ${expr.op} ${val};`);
        this.releaseVars(val);
        return code.join('');
    }
    static isArrayFilterExpr(expr) {
        return expr.type === types_1.SyntaxType.ARRAY_FILTER_EXPR;
    }
    static isValidSelectorForAssignment(expr) {
        return expr.selector === '.' && !!expr.prop && expr.prop.type !== types_1.TokenType.PUNCT;
    }
    static isSinglePropSelection(expr) {
        if (expr.type === types_1.SyntaxType.SELECTOR) {
            const part = expr;
            return (part.selector === '.' &&
                (part.prop?.type === types_1.TokenType.ID || part.prop?.type === types_1.TokenType.STR));
        }
        return false;
    }
    static returnObjectValues(varName) {
        return `Object.values(${varName}).filter(v => v !== null && v !== undefined)`;
    }
    static returnSingleValue(arg, varName) {
        if (arg.type === types_1.SyntaxType.LITERAL) {
            return varName;
        }
        return `(Array.isArray(${varName}) ? ${varName}[0] : ${varName})`;
    }
    static convertToSingleValue(varName) {
        return `${varName} = Array.isArray(${varName}) ? ${varName}[0] : ${varName};`;
    }
    static returnSingleValueIfSafe(varName) {
        return `(${varName}.length < 2 ? ${varName}[0] : ${varName})`;
    }
    static covertToArrayValue(varName) {
        return `${varName} = Array.isArray(${varName}) ? ${varName} : [${varName}];`;
    }
    static generateAssignmentCode(key, val) {
        return `${key}=${val};`;
    }
    translateArrayFilterExpr(expr, dest, ctx) {
        const code = [];
        code.push(JsonTemplateTranslator.generateAssignmentCode(dest, ctx));
        for (const filter of expr.filters) {
            if (filter.type === types_1.SyntaxType.ARRAY_INDEX_FILTER_EXPR) {
                code.push(this.translateIndexFilterExpr(filter, dest, dest));
            }
            else {
                code.push(this.translateRangeFilterExpr(filter, dest, dest));
            }
        }
        return code.join('');
    }
    translateObjectFilterExpr(expr, dest, ctx) {
        const code = [];
        const condition = this.acquireVar();
        code.push(this.translateExpr(expr.filter, condition, ctx));
        code.push(`if(!${condition}) {${dest} = undefined;}`);
        this.releaseVars(condition);
        return code.join('');
    }
    translateObjectIndexFilterExpr(ctx, allKeys, resultVar, shouldExclude) {
        const code = [];
        if (shouldExclude) {
            code.push(`${allKeys}=Object.keys(${ctx}).filter(key => !${allKeys}.includes(key));`);
        }
        code.push(JsonTemplateTranslator.generateAssignmentCode(resultVar, '{}'));
        code.push(`for(let key of ${allKeys}){`);
        code.push(`if(Object.prototype.hasOwnProperty.call(${ctx}, key)){${resultVar}[key] = ${ctx}[key];}`);
        code.push('}');
        return code.join('');
    }
    translateArrayIndexFilterExpr(ctx, allKeys, resultVar) {
        const code = [];
        code.push(JsonTemplateTranslator.generateAssignmentCode(resultVar, '[]'));
        code.push(`for(let key of ${allKeys}){`);
        code.push(`if(typeof key === 'string'){`);
        code.push(`for(let childCtx of ${ctx}){`);
        code.push(`if(Object.prototype.hasOwnProperty.call(childCtx, key)){`);
        code.push(`${resultVar}.push(childCtx[key]);`);
        code.push('}');
        code.push('}');
        code.push('} else {');
        code.push(`if(key < 0){key = ${ctx}.length + key;}`);
        code.push(`if(Object.prototype.hasOwnProperty.call(${ctx}, key)){`);
        code.push(`${resultVar}.push(${ctx}[key]);`);
        code.push('}');
        code.push('}');
        code.push('}');
        code.push(`if(${allKeys}.length === 1) {${resultVar} = ${resultVar}[0];}`);
        return code.join('');
    }
    translateIndexFilterExpr(expr, dest, ctx) {
        const code = [];
        const allKeys = this.acquireVar();
        code.push(this.translateArrayExpr(expr.indexes, allKeys, ctx));
        code.push(`${allKeys} = ${allKeys}.flat();`);
        const resultVar = this.acquireVar();
        if (expr.type === types_1.SyntaxType.OBJECT_INDEX_FILTER_EXPR) {
            code.push(this.translateObjectIndexFilterExpr(ctx, allKeys, resultVar, expr.exclude));
        }
        else {
            code.push(this.translateArrayIndexFilterExpr(ctx, allKeys, resultVar));
        }
        code.push(JsonTemplateTranslator.generateAssignmentCode(dest, resultVar));
        this.releaseVars(allKeys);
        this.releaseVars(resultVar);
        return code.join('');
    }
    translateRangeFilterExpr(expr, dest, ctx) {
        const code = [];
        let fromIdx = this.acquireVar();
        let toIdx = this.acquireVar();
        if (expr.fromIdx) {
            if (expr.toIdx) {
                code.push(this.translateExpr(expr.fromIdx, fromIdx, ctx));
                code.push(this.translateExpr(expr.toIdx, toIdx, ctx));
                code.push(dest, '=', ctx, '.slice(', fromIdx, ',', toIdx, ');');
            }
            else {
                code.push(this.translateExpr(expr.fromIdx, fromIdx, ctx));
                code.push(dest, '=', ctx, '.slice(', fromIdx, ');');
            }
        }
        else if (expr.toIdx) {
            code.push(this.translateExpr(expr.toIdx, toIdx, ctx));
            code.push(dest, '=', ctx, '.slice(0,', toIdx, ');');
        }
        this.releaseVars(fromIdx, toIdx);
        return code.join('');
    }
    translateBinaryExpr(expr, dest, ctx) {
        const val1 = this.acquireVar();
        const val2 = this.acquireVar();
        const { args } = expr;
        const code = [];
        code.push(this.translateExpr(args[0], val1, ctx));
        code.push(this.translateExpr(args[1], val2, ctx));
        code.push(dest, '=', operators_1.binaryOperators[expr.op](JsonTemplateTranslator.returnSingleValue(args[0], val1), JsonTemplateTranslator.returnSingleValue(args[1], val2)), ';');
        this.releaseVars(val1, val2);
        return code.join('');
    }
}
exports.JsonTemplateTranslator = JsonTemplateTranslator;
